/* 
 *  blood sky heart
 *  Entry to povcomp 2005
 *  Copyright Simeon Scott, 31 January 2005
 *  email shevek at bur.st
 *  
 *  Unfortunately, povray is as much an art as a science, and so by
 *  the time any scene (of mine) becomes aesthetic, its code has generally
 *  become ugly. My apologies for any excess difficulty in understanding
 *  the code.
 *  
 *  I have adapted a procedure which I created for another purpose:
 *  filling a space with tons of bubbles and without intersecting.
 *  
 *  Here most of that effort may not be needed, since blobs automatically
 *  take care of what happens at intersection. But since I have that
 *  procedure already, it's more efficient just to use it, than to
 *  rewrite it. I didn't quite have time to make the procedure fully
 *  generic, so that either blobs or 'bubbles' could be produced.
 *  
 *  An include file is generated defining the positions of blobs. This
 *  means subsequent renders can be done a bit faster, and you can 
 *  hand-tune the locations of individual bloblets. Textures are defined
 *  in the main file, or you could bring them in using other #include files.
 *  
 *  If using a great number of bubbles, it may take quite a long time to 
 *  generate the bubbles include file.
 *  If you interrupt generation using CTRL-C or similar, you should be
 *  able to use the include file anyway, with only minor manual edits. 
 *  The bubbles are generated in no particular order, but if it looks 
 *  wildly asymetric, it's because there are simply too few bubbles. 
 *  Statistically they will balance out one there are several hundred.
 *  
 *  Rendering the scene a second time when there is an existing
 *  include file (even if it's empty) means the existing include file
 *  will be used, none will be generated. Look out!
 *  
 *  When incrementing version numbers of file, remember to appropriately 
 *  increment version number variable Our_Version so that the correct
 *  include file is used/generated.
 *  
 *  Rendered for gamma=1.0 colourspace, using unix PoVRay:
 *
 *  povray +A0.3 +AM2 +R3 +GD +P Display_Gamma=1.0 blood_sky_heart.pov +W1024 +H1200 +Oblood_sky_heart_main.png
 *
 *  Zoom sections generated as per povcomp.com rules, suitable for 
 *  conversion using the povcomp-supplied partial-PPM-render-to-PNG
 *  convertor script, use:
 *  
 *  povray +GD +P +A0.3 +AM2 +R3 +H4688 +W4000 Display_Gamma=1.0 blood_sky_heart.pov +SR0.1667 +ER0.3667 +SC0.4004 +EC0.6006 +FP +Oblood_sky_heart_zoom1.ppm
 *  
 *  povray +GD +P +A0.3 +AM2 +R3 +H4688 +W4000 Display_Gamma=1.0 blood_sky_heart.pov +SR0.6883 +ER0.8833 +SC0.7999 +EC1 +FP +Oblood_sky_heart__zoom2.ppm
 *
 *	The two resulting PPM files can then be converted on unix, using the
 *	povcomp supplied convertor script with the following commands:
 *
 *	sh img_crop.sh blood_sky_heart_zoom1.ppm 1601 2402 781 1719
 *
 * 	sh img_crop.sh blood_sky_heart_zoom2.ppm 3199 4000 3226 4140
 *
 */

/* 
 * Set parameters here. 
 */
#declare Our_Version = "blood_sky_heart"	// used in include file name
//#declare Preview = 1;					// comment out for full quality
#declare Seed_Num = 71;					// randomiser
#declare Count = 700;					// How many bubbles?
#declare Size_Min = 0.01;				// size range of bubbles
#declare Size_Max = 0.04;
#declare Gap_Min = 0.04;				// min space between bubbles
/*
 *	Loops_Max is a bail-out. I find about 10000 tries can get you 
 *	500 mid-size bubbles. 100000 tries gets you about 1000 of them.
 *	This many attempts is quite feasible on a 1GHz PIII
 *	and smaller Size_Min fits them in more easily (ie with fewer tries).
 */
#declare Loops_Max = Count * 15;	// Bail out after how many attempts?

/* 
 * End of tweaks section. Change anything below at your own risk. 
 */

global_settings 
{ 
	assumed_gamma 1.0 
	max_trace_level 12

	#ifndef ( Preview )
		photons
		{		
			media 12
			spacing 0.005
			autostop 0
			jitter 0.4
			radius
		}
	#end
}

#declare Pos = array[Count]
#declare Size = array[Count]
#declare Seed = seed(Seed_Num);
#declare Generate_File = concat ( Our_Version, ".inc" )

/*
 * Temporary variable so that maths aren't
 *	carried out every time a loop is tested.
 */
#declare Space_Min = Size_Min + Gap_Min;

/*
 * In order to write out a string later describing the
 *	texture to use, declare a string here naming the texture,
 *	then below, declare the texture which matches the string. 
 */
#ifndef ( Preview )
	#declare Bubble_Texture1 = "Transfuse_Red"
	#declare Bubble_Interior1 = "Transfuse_Interior1"
#else
	#declare Bubble_Texture1 = "Flat_Red"
	#declare Bubble_Interior1 = "Default_Interior"
#end

#declare Flat_Red = texture 			// faster for preview
{ 
	pigment { color rgb < 1, 0, 0 >  }
	finish 
	{ 
		ambient 0
		diffuse 0.6
		specular 1 
		roughness 0.01
	}

}
#declare Default_Interior = interior {}	// faster for preview

#declare Transfuse_Red = texture
{ 
	
	pigment { color rgbt < 1, 0.2, 0, 1 >  }
	finish 
	{ 
		ambient 0
		diffuse 0 
		brilliance 3
		reflection { 
			fresnel on
			metallic 2
		}
		specular 1
		roughness 0.007	
		conserve_energy
	}
	normal 
	{
		ripples scale 0.025 
		turbulence 0.4 
	}
}

#declare Transfuse_Interior1 = interior 
{ 
/* 
 *	Saline water ior from
 *	http://www-public.tu-bs.de:8080/~y0013390/pov/water/water04.html
 *	The following two references not used 'scientifically', but
 *	here for interest's sake.
 *	Blood scattering described in
 *	http://www-atom.fysik.lth.se/AFDOCS/Progrep956/4b.htm
 *	Blood optical properties at
 *	http://omlc.bme.ogi.edu/spectra/hemoglobin/index.html
 */
	ior 1.34  //saline water
	fade_distance 0.015
	fade_power 100000 	// over 1000 sets use of exponential fade
	media
	{
        intervals 1         
         samples 1,1          
		absorption rgb < 0.1, 1, 1 >
		scattering { 1, rgb < 0.7, 0.06, 0 > }
		emission < 0.45, 0, 0 >	// to simulate ambient light transmission
	}
}

/*
 * heart isosurface function from 
 *	http://atrey.karlin.mff.cuni.cz/~0rfelyus/povray/index.html
 *	cub ( 2 * sqr(x) + sqr(y) + sqr(z) -1 )
 *	 - ( 0.1 * sqr(x) + sqr(y) ) * cub(z)
 *
 *	or in POVRay 3.6 language:
 *
 *	pow ( ( 2 * pow ( x, 2 ) + pow ( y, 2 ) + pow ( z, 2 ) -1 ), 3 )
 *	 - ( 0.1 * pow ( x, 2 ) + pow (y, 2 ) ) * pow ( z, 3 )
 *
 *	I use it rotated facing camera in -z: swap x, y, z <=> z, x, y
 *
 *	For interest, see also parametric version
 *	http://www.uib.no/People/nfytn/mathgal.htm
 *	
 *	Function also available now in builtins using functions.inc,
 *	although it doesn't seem any faster or more accurate. I typed it
 * 	out in full again, for the reader's interest.
 *	
 *	Excellent POVRay isosurface information here:
 *	http://www.econym.demon.co.uk/isotut/
 *	
 *	Tried also, but to no avail:
 *	http://news.povray.org/povray.binaries.scene-files/attachment/%3C3f088ef1%241%40news.povray.org%3E/isosurfaceJF.zip
 */
#declare Heart_Function = function (x, y, z) {
	 pow ( ( 2 * pow ( z, 2 ) + pow ( x, 2 ) + pow ( y, 2 ) -1 ), 3 )
	 - ( 0.1 * pow ( z, 2 ) + pow (x, 2 ) ) * pow ( y, 3 )
}

/* 
 *	Describe the container shape here which will be
 *	filled with bubbles. 
 */
#declare Container = isosurface 
{ 
	function { Heart_Function (x, y, z) }

/*
 *	Accuracy:
 *	Render times shown for rendering the object, and how it looks.
 *
 * 	one container object accuracy 0.001 and big ugly scar artifact
 * 	Render Time:   0 hours  9 minutes  5 seconds (545 seconds)
 *
 * 	one container object accuracy 0.000001 max_gradient 36 max_trace 1
 * 	no scar
 * 	Render Time:   3 hours 14 minutes 36 seconds (11676 seconds)
 *
 *	In fact in this case it is less important how it looks, since
 *	we are just using it as a container using 'inside' tests.
 *	This also seems not to be very slow, so I used a high accuracy
 *	'just in case'.
 */
	accuracy 0.000001  	// about perfect
//	accuracy 0.001			// ugly horizontal scar
//	accuracy 0.01   		// misshapen
	max_trace 1				// if solid
//	max_trace 4				// if transparent/open
//	max_gradient 230.2		// reported (approximately)
//	max_gradient 150		// seems always clean
	max_gradient 36			// seems always clean
	/*
	 *	Default bounding is a unit sphere (or box?)
	 *	sphere { < 0, 0, 0 >, 1 }
	 *	which would truncate the surface.
	 *	This occurs before any scaling, so must be fixed here
	 *	(unless you figure out the maths to do it earlier 
	 *	in the function declaration). 
	 */
	contained_by { box { < 1.17, 1.25, 0.74 >, -< 1.17, 1.25, 0.74 > } }
	// fit it into unit sphere
	scale 0.8
	translate < 0, -0.2, 0 >
}
// For quicker debugging, can use some other object instead.
//#declare Container = sphere { < 0, 0, 0 >, 1 texture { Our_Texture1} }

/*
 * 	A hole, culls AFTER the bubble is generated, so that
 *	we can spped up rendering of particular views, while
 *	also being able to reproduce the entire object with identical
 *	bubbles matching the ones we used in the culled version.
 *	Used in this case to knock a whopping great hole in the middle,
 *	for visual effect.
 */
#declare Hole = union 
{ 
	sphere { < 0, -0.1, 0.55 >, 0.55 }
	sphere { < 0, -0.47, 0.4 >, 0.4 }
	sphere { < -0.33, 0.22, 0.5 >, 0.5 }
	sphere { < 0.33, 0.22, 0.5 >, 0.5 }
	sphere { < 0, -0.1, -0.55 >, 0.55 }
	sphere { < 0, -0.47, -0.4 >, 0.4 }
	sphere { < -0.33, 0.22, -0.5 >, 0.5 }
	sphere { < 0.33, 0.22, -0.5 >, 0.5 }
	scale < 0.7, 0.7, 1 >
}

/*
 *	Process below inspired by	
 *	FillCylinder by Friedemann Schmidt webmaster@friedemann-schmidt.de
 *	(no longer available online) 
 *	Raytracing-Gallery http://www.friedemann-schmidt.de
 *
 *	Loop over all existing objects (approximated as a sphere)
 *	placed thus far, checking if the new object (ditto)
 *	intersects with any of the existing ones.
 *
 *	This is not really necessary when using blobs, but it is used anyway,
 *	as explained in comments at top of file.
 */

//TODO make seperate macros, Write_Blob_Component vs Write_Bubble
#macro Write_Object (Pos_Write, Size_Write)
	#debug concat( "writing blob component #", str( Num_Bubbles, 0, 0 ), "\n" )
	
	#fopen out Generate_File append
		#write ( out, concat ( 
			"sphere { < 0, 0, 0 >, 3, 2 scale ", str( Size_Write, 0, -1 ), " ",
			"translate < ", vstr( 3, Pos_Write, ", ", 0, -1 ), " > ",
			"}\n" ) )
	#fclose out
#end

#macro Try_Add_Object ( Pos_Temp )
	#declare Wrote_One = 0;		// flag to pass back to caller

	#if ( Num_Bubbles = 0 )		// ... is this our first bubble?

		/* TODO Move to a macro so that user can set 
			a non-linear random function to be used. */
		#declare Size_Temp = Size_Max * rand(Seed);

		Write_Object ( Pos_Temp, Size_Temp )
		#declare Wrote_One = 1;
		#declare Pos[0] = Pos_Temp;
		#declare Size[0] = Size_Temp;

	#else
		/*
		 * 	See if the proposed bubble is too close to existing bubbles.
		 *
		 *	We also keep track of the smallest distance found,
		 *	which may determine the size of the bubble. 
		 */

		/*
		 * 	If Size_Temp is not yet defined, we simply set it 
		 *	using the first bubble in the list. 
		 */
		#local Space_Temp = vlength ( Pos_Temp - Pos[0] ) - Size[0];
		#local Size_Temp = Space_Temp - Gap_Min;
		/*
		 *	Use temporary variable so that maths not 
		 *	done in testing loop condition. 
		 */
		#local Space_Best_So_Far = Space_Temp;
		
		#local j = 1;				// counter
		#local Try = 1;				// bail-out boolean

		#while ( Try & j < Num_Bubbles )

			#local Space_Temp = vlength ( Pos_Temp - Pos[j] ) - Size[j];

			#if ( Space_Temp >= Space_Min )

				/*
				 * 	Compare and see if the space to the next bubble 
				 *	is less than tracking variable Size_Temp and 
				 *	reset Size_Temp if so. 
				 */
				#if ( Space_Temp < Space_Best_So_Far )
					#local Size_Temp = Space_Temp - Gap_Min;
				/*
				 * 	Use temporary variable so that maths not 
				 *	done in testing loop condition. 
				 */
					#local Space_Best_So_Far = Space_Temp;
				#end
				#local j = j + 1;
			#else
				#local Try = 0;
			#end

		#end

		#if ( Try )
			// If bubble is larger than max, resize it.
			#if ( Size_Temp > Size_Max )

				/* TODO Move to a macro so that user can set 
					a non-linear random function to be used. */
				#local Size_Temp = Size_Max * rand(Seed);

			#end

			// Keep the bubble only if the space is large enough.
			#if ( Size_Temp > Size_Min )
				Write_Object ( Pos_Temp, Size_Temp )
				#declare Wrote_One = 1;
				/*
				 * Also add the new bubble to our test lists. 
				 *	Remember that j has already been incremented 
				 *	during the loop. 
				 */
				#declare Pos[j] = Pos_Temp;
				#declare Size[j] = Size_Temp;
			#end

		#end
	#end
#end

#macro Bubbles ( Container_to_Use, Loops_Max)
	/*
	 *	Fill Pos and Size arrays with bubble positions and sizes
	 *	where no bubble intersects with another bubble, and all bubbles
	 *	are contained within a given object.
	 */

	#declare Loops_Num = 0;
	/* 
	 *	If too many unsuccessful tries, bail out.
	 *	If enough bubbles found, then bail out.
	 *	Remember test is done before loop is entered so values
	 *	incremented inside the loop are detected only next 
	 *	iteration. 
	 */
	#while ( Num_Bubbles < Count & Loops_Num < Loops_Max )
		
		// Choose random Pos_New.
		#local Pos_New = 
			< 1 - rand(Seed) * 2, 1 - rand(Seed) * 2, 1 - rand(Seed) * 2 >;

		// See if Pos_New is within the container object...
	
		#if ( ! inside ( Hole, Pos_New ) )
			#if ( inside ( Container, Pos_New ) )
				Try_Add_Object ( Pos_New )
				#if ( Wrote_One )
					// Just added an item and wrote to the include file.
					#declare Num_Bubbles = Num_Bubbles + 1;
				#end
			#end
		#end

	#declare Loops_Num = Loops_Num + 1;
	#end
	/*
	 * Note for debug output: Loops_Num has been incremented, but also
	 *	remember that the first loop number was 0, so Loops_Num now
	 *	actually contains the number of loops so far counted from 1. :) 
	 */
	#warning concat( 
		"Number of bubbles written ", str( Num_Bubbles, 0, 0 ), 
		"\nNumber of attempts in last loop was ", str( Loops_Num, 0, 0 ),
		"\n" )
#end

/*
 * 	Now use the above macros to generate the include file which
 *	lists the positions of the many sub-blobs.
 */

// If include file exists, use it, else create a new one first.
#if ( file_exists ( Generate_File ) )
	#warning concat (
			"Using existing include file, ", Generate_File, "\n"
		)
	#include Generate_File
#else

	// Initialise a new include output file.
	#fopen out Generate_File write
		#write ( out, concat ( 
			"// ", Generate_File, " autogenerated  by ", Our_Version, "\n",
			"#declare Conglomerate = blob {\n"
		) )
	#fclose out

	#declare Num_Bubbles = 0;
	Bubbles ( Container, Loops_Max)

	// Close blob object.
	#fopen out Generate_File append
		#write ( out, concat ( 
			"texture { ", Bubble_Texture1, "}\n",
			"interior { ", Bubble_Interior1, "}\n",
			"}\n"
		) )
	#fclose out

	#if ( file_exists ( Generate_File ) )
		#include Generate_File
	#else
		#warning concat (
			"Generated the file, ", Generate_File, " but now can't find it!\n"
		)
	#end
#end


/*
 *	Now set up the rest of the scene.
 */

// Inner ground
sphere
{ 
	< 0, -1018, 0 >, 1000
	texture
	{
		pigment 
		{
			crackle
			scale 1.5
	    	color_map 
			{
				[ 0	color rgb < 0.6, 0.7, 0.8 > * 0.1 ]
				[ 0.0005	color rgb < 0.6, 0.7, 0.8 > * 0.1 ]
				[ 0.6	color rgb < 0.8, 0.7, 0.6 > ]
				[ 0.8	color rgb < 0.6, 0.8, 0.6 > ]
			}
		}
	}
	normal
	{
		crackle 1
		scale 1.5
	}
}

// Outer ground
sphere
{ 
	< 0, -1017.999999, 0 >, 1000
	texture	
	{
		pigment 
		{
     		 function{abs(x*x + z*z)}
	  
      		color_map 
	  		{
				[ 0		color rgbt < 0.9, 0.8, 0.8,  1 > ]
				[ 0.4		color rgbt < 0.9, 0.8, 0.8, 0 > ]
			}
		}
		scale 150
		translate x+3
	}				
	normal
	{
		crackle 0.2
		scale 3
	}
}

/*
 *	Some hills using a sphere function with noise function added
 *  see http://www.econym.demon.co.uk/isotut/combine.htm
 * 
 */
#include "functions.inc"
isosurface 
{
	function 
	{
		( 1.15*x*x + 1.15*y*y + 1.15*z*z - 1 )		// small sphere
		- f_noise3d(x*25, y*25, z*25)				// hills with
		 * ( x*x + z*z ) * 0.15						// a plane at x,z = 0
		+ f_noise3d(x*350, y*350, z*350) * 0.01		// bumpy (indented)
	}
	max_gradient 4.7
	contained_by{sphere{0,1}}
	rotate y*38
	rotate x*-37
	rotate y*6
	rotate x*5
	rotate y*25

	scale 1060
	translate y*-1018						// align with crusty sphere

	texture
	{
		pigment 
		{
			crackle 
			scale 4.5
			turbulence 0.6
	    	color_map 
			{
				[ 0	color rgb < 0.6, 0.8, 0.8 > ]
				[ 0.0005	color rgb < 0.8, 0.7, 0.6 > ]
				[ 0.8	color rgb < 0.6, 0.7, 0.8 > * 0.1 ]
				[ 1	color rgb < 0.6, 0.7, 0.8 > * 0.1 ]
			}
		}
	}
/*	normal
	{
		crackle 1
		scale 6
	}*/
}

/*
 *	Slight hazing at horizon - 'fog' feature is no good,
 *	as ground is curved away from us.
 */
#ifndef ( Preview )
difference
{
	sphere { < 0, -540, 0 >, 540 }
	sphere { < 0, -517, 0 >, 500 }
	no_shadow
	hollow on
	texture { pigment { color rgbt < 0, 0, 0, 1 > } }
	interior 
	{ 
		ior 1  //air - default
		media
		{
			density { function { (x*x + y*y + z*z - 500) / 40} }
			method 3
       		intervals 1
			jitter 1       
	         samples 10
			aa_threshold  0.02  
			emission < 0.8, 0.8, 1 > * 0.004
		}
	}
}
#end


// Interesting sky
background { rgb 0 }
sky_sphere 
{
    pigment 
	{
      function{abs(y + 3/2*x*x)}
	  
      color_map 
	  {
		[ 0		color rgb < 0.8, 0.8, 1 > ]
        [ 0.5 	color rgb < 0.4, 0.4, 1 > ]
        [ 0.94 	color rgb < 0, 0, 0.1 > ]
      }

	/* 
	 *	Documentation says sky_sphere is independent
	 *	of camera ? .. but this is not the case. ?
	 */
	translate < 0, 0, -4 >
	rotate x*15
	rotate y*-10
	}
}

/*
// To draw the container itself for debugging, uncomment this section.
object { Container texture texture { pigment { color rgbf < 1, 0, 0, 0.3 > } }
//	rotate y*90
}*/

// To draw Bubble for debugging, uncomment below.
// object { Bubble	texture { pigment { color rgbt < 1, 0, 1, 0.8 > } }  }

// To draw unit sphere for debugging, uncomment below.
//sphere { < 0, 0, 0 >, 1  pigment { color rgbt < 1, 1, 1, 0.8 > }}

/* 
// To draw our optimisation slabs used in the container object,
// uncomment this section.
box { < 0.75, 1.17, 1.25 >, < -0.75, -1.17, Tricky_Bound_Top >
	texture { pigment { color rgbt < 1, 0, 1, 0.95 > } }
	scale 0.8
	rotate < 0, 90, 0 >
	rotate < 0, 0, 90 >
	translate < 0, -0.2, 0 >
}
box { < 0.74, 1.05, Tricky_Bound_Top > , < -0.74, -1.05, Tricky_Bound_Bot > 
	texture { pigment { color rgbt < 0, 1, 0, 0.95 > } }
	scale 0.8
	rotate < 0, 90, 0 >
	rotate < 0, 0, 90 >
	translate < 0, -0.2, 0 >
} 

box { < 0.72, 1, Tricky_Bound_Bot >, < -0.72, -1, -1.25 >
	texture { pigment { color rgbt < 1, 1, 0, 0.95 > } }
	scale 0.8
	rotate < 0, 90, 0 >
	rotate < 0, 0, 90 >
	translate < 0, -0.2, 0 >
 } */

// To draw the culling object for debugging, uncomment below.
// object { Hole texture { pigment { color rgbt < 1, 1, 1, 0 > } } }

/*
 * The heart/blob object.
 */
object { Conglomerate 
	hollow 
	photons
	{
		target 0.5 // spacing multiplier
		reflection on
		refraction on
		collect off
	}
	translate < -0.9, 2.6, 4.8 >
}

/* 
 *	When adjusting lighting, remember that photons switched off
 *	makes a big difference.
 */


light_source {
	< 0, 0, -8 > * 100000
	color rgb 1
	media_attenuation on
	
	#ifndef ( Preview )
//		area_light x*3, y*3, 7, 7
//		adaptive 2
//		circular
	#end
	rotate x * 26				// gives elevation
	rotate y * -60				// gives position

	photons
	{
//		area_light
		refraction on
		reflection on
	}
//	rotate y*-90		// side view
}

// For highlights
light_source {
	< 0, 0, -8 >
	color rgb 1
	media_attenuation on
	rotate x * 26				// gives elevation
	rotate y * 120				// gives position
	translate < -0.9, 2.6, 4.8 >
	photons
	{
		refraction on
		reflection on
	}
}

camera {
//	orthographic
	up 1200/1024 * y// * 2.55
	right 1 * x// * 2.55
	location < 0, 0, -8 >	// long lens if using perspective 
    look_at  < 0, 0,  0 >

	// position camera slightly off-centre
	rotate x*-5
	rotate y*-10
	translate -1.5*y
	translate -0.1*x
	
//	rotate y*90

//	rotate y*90		// side view

//	rotate x*-90		// top view
//	rotate y*180

//	rotate x*15		// floating - use with perspective
//	rotate y*30
}
